import unittest

from deep_phonemizer.training.decorators import ignore_exception


@ignore_exception
def test_function() -> None:
    raise ValueError("Exception should be ignored.")


class TestDecorators(unittest.TestCase):

    def test_ignore_exception(self):
        test_function()



import math
from typing import Any, List, Union


def _product(probs: Union[None, List[float]]) -> float:
    if probs is None or len(probs) == 0:
        return 0.0
    if 0 in probs:
        return 0
    prob = math.exp(sum([math.log(p) for p in probs]))
    return prob


def _batchify(input: List[Any], batch_size: int) -> List[List[Any]]:
    l = len(input)
    output = []
    for i in range(0, l, batch_size):
        batch = input[i : min(i + batch_size, l)]
        output.append(batch)
    return output



import unittest

from deep_phonemizer.preprocessing.text import LanguageTokenizer


class TestLanguageTokenizer(unittest.TestCase):

    def test_call_happy_path(self) -> None:

        tokenizer = LanguageTokenizer(languages=["de", "en"])

        de_tok = tokenizer("de")
        en_tok = tokenizer("en")
        self.assertEqual(0, de_tok)
        self.assertEqual(1, en_tok)

        self.assertEqual("de", tokenizer.decode(de_tok))
        self.assertEqual("en", tokenizer.decode(en_tok))

    def test_exception(self) -> None:
        tokenizer = LanguageTokenizer(languages=["de", "en"])

        with self.assertRaises(ValueError):
            tokenizer("not_existent")



import os
import shutil
import tempfile
import unittest
from pathlib import Path

from deep_phonemizer import preprocess
from deep_phonemizer.preprocess import preprocess
from deep_phonemizer.preprocessing.text import Preprocessor
from deep_phonemizer.utils.io import read_config, save_config, unpickle_binary


class TestPreprocess(unittest.TestCase):

    def setUp(self) -> None:
        temp_dir = tempfile.mkdtemp(prefix="TestPreprocessTmp")
        self.temp_dir = Path(temp_dir)
        test_path = os.path.dirname(os.path.abspath(__file__))
        self.test_config_path = Path(test_path) / "resources/forward_test_config.yaml"

    def tearDown(self) -> None:
        shutil.rmtree(self.temp_dir, ignore_errors=True)

    def test_preprocess_happy_path(self) -> None:
        config = read_config(self.test_config_path)
        data_dir = self.temp_dir / "datasets"
        config_path = self.temp_dir / "forward_test_config.yaml"
        config["paths"]["data_dir"] = str(data_dir)
        save_config(config, config_path)

        train_data = [
            ("de", "benützten", "bəˈnʏt͡stn̩"),
            ("de", "gewürz", "ɡəˈvʏʁt͡s"),
            ("en_us", "young", "jʌŋ"),
        ]

        val_data = [("en_us", "young", "jʌŋ"), ("de", "gewürz", "ɡəˈvʏʁt͡s")]

        preprocess(
            config_file=config_path,
            train_data=train_data,
            val_data=val_data,
            deduplicate_train_data=False,
        )

        preprocessor = Preprocessor.from_config(config)

        expected_train = [preprocessor(t) for t in train_data]
        actual_train = unpickle_binary(data_dir / "train_dataset.pkl")
        self.assertEqual(expected_train, actual_train)

        expected_val = [preprocessor(t) for t in val_data]
        actual_val = unpickle_binary(data_dir / "val_dataset.pkl")
        self.assertEqual(expected_val, actual_val)

        expected_phon_dict = {"de": {}, "en_us": {}}
        for lang, word, phon in train_data + val_data:
            expected_phon_dict[lang][word] = phon

        actual_phon_dict = unpickle_binary(data_dir / "phoneme_dict.pkl")
        self.assertEqual(expected_phon_dict, actual_phon_dict)



import math
from collections import Counter
from pathlib import Path
from typing import Any, Dict, List, Tuple

import torch
import tqdm
from torch.nn.parallel import DistributedDataParallel
from torch.optim import Adam
from torch.optim.lr_scheduler import ReduceLROnPlateau
from torch.utils.tensorboard import SummaryWriter

from deep_phonemizer.model.model import Model
from deep_phonemizer.model.utils import _trim_util_stop
from deep_phonemizer.preprocessing.text import Preprocessor
from deep_phonemizer.training.dataset import new_dataloader
from deep_phonemizer.training.decorators import ignore_exception
from deep_phonemizer.training.evaluation import evaluate_samples
from deep_phonemizer.training.losses import CrossEntropyLoss, CTCLoss
from deep_phonemizer.utils.io import to_device, unpickle_binary


class Trainer:
    """Performs model training."""

    def __init__(
        self,
        checkpoint_dir: Path,
        device: torch.device,
        rank: int,
        use_ddeep_phonemizer: bool,
        loss_type="ctc",
    ) -> None:
        """
        Initializes a Trainer object.

        Args:
          checkpoint_dir (Path): Directory to store the model checkpoints.
          device (torch.device): Device used for training
          rank (int): Rank of the current device
          use_ddeep_phonemizer (bool): Flag whether Ddeep_phonemizer is used for training
          loss_type (str): Type of loss: 'ctc' for forward transformer models
                           and 'cross_entropy' for autoregressive models.
        """

        self.checkpoint_dir = checkpoint_dir
        self.checkpoint_dir.mkdir(parents=True, exist_ok=True)
        self.use_ddeep_phonemizer = use_ddeep_phonemizer
        self.rank = rank
        self.device = device
        self.writer = SummaryWriter(log_dir=str(self.checkpoint_dir / "logs"))
        self.loss_type = loss_type
        if loss_type == "ctc":
            self.criterion = CTCLoss()
        elif loss_type == "cross_entropy":
            self.criterion = CrossEntropyLoss()
        else:
            raise ValueError(f"Loss not supported: {loss_type}")

    def train(
        self,
        model: Model,
        checkpoint: Dict[str, Any],
        store_phoneme_dict_in_model: bool = True,
    ) -> None:
        """
        Performs training of a transformer model.

        Args:
          model (Model): Model to be trained (can be a fresh model or restored from a checkpoint).
          checkpoint (Dict[str, Any]): Dictionary with entries 'optimizer': optimizer state dict,
                                       'preprocessor': Preprocessor and 'config': Dict.
          store_phoneme_dict_in_model (bool): Whether to store a dictionary of word-phoneme mappings
                                              in the model checkpoint so that it can be automatically
                                              loaded by a Phonemizer object.

        Returns:
          None: the checkpoints will be stored in a folder provided when instantiating a Trainer.
        """

        config = checkpoint["config"]
        data_dir = Path(config["paths"]["data_dir"])

        model = model.to(self.device)
        model.train()

        if self.use_ddeep_phonemizer:
            model = DistributedDataParallel(model, device_ids=[self.rank])

        criterion = self.criterion.to(self.device)

        optimizer = Adam(model.parameters())
        if "optimizer" in checkpoint:
            optimizer.load_state_dict(checkpoint["optimizer"])
        for g in optimizer.param_groups:
            g["lr"] = config["training"]["learning_rate"]

        train_loader = new_dataloader(
            dataset_file=data_dir / "train_dataset.pkl",
            use_ddeep_phonemizer=self.use_ddeep_phonemizer,
            drop_last=True,
            batch_size=config["training"]["batch_size"],
        )

        if self.rank == 0:
            val_loader = new_dataloader(
                dataset_file=data_dir / "val_dataset.pkl",
                use_ddeep_phonemizer=False,
                drop_last=False,
                batch_size=config["training"]["batch_size_val"],
            )

            val_batches = sorted(
                [b for b in val_loader], key=lambda x: -x["text_len"][0]
            )

        if store_phoneme_dict_in_model:
            phoneme_dict = unpickle_binary(data_dir / "phoneme_dict.pkl")
            checkpoint["phoneme_dict"] = phoneme_dict

        scheduler = ReduceLROnPlateau(
            optimizer,
            factor=config["training"]["scheduler_plateau_factor"],
            patience=config["training"]["scheduler_plateau_patience"],
            mode="min",
        )
        losses = []
        best_per = math.inf
        if "step" not in checkpoint:
            checkpoint["step"] = 0
        start_epoch = checkpoint["step"] // len(train_loader)

        for epoch in range(start_epoch + 1, config["training"]["epochs"] + 1):
            if self.use_ddeep_phonemizer:
                train_loader.sampler.set_epoch(epoch)
            pbar = tqdm.tqdm(enumerate(train_loader, 1), total=len(train_loader))
            for i, batch in pbar:
                checkpoint["step"] += 1
                step = checkpoint["step"]
                self._set_warmup_lr(optimizer=optimizer, step=step, config=config)
                batch = to_device(batch, self.device)
                avg_loss = sum(losses) / len(losses) if len(losses) > 0 else math.inf
                pbar.set_description(
                    desc=f"Rank: {self.rank} | Epoch: {epoch} | Step {step} "
                    f"| Loss: {avg_loss:#.4}",
                    refresh=True,
                )

                pred = model(batch)
                loss = criterion(pred, batch)

                if not (torch.isnan(loss) or torch.isinf(loss)):
                    optimizer.zero_grad()
                    loss.backward()
                    torch.nn.utils.clip_grad_norm_(model.parameters(), 1.0)
                    optimizer.step()
                    losses.append(loss.item())

                self.writer.add_scalar("Loss/train", loss.item(), global_step=step)
                self.writer.add_scalar(
                    "Params/batch_size",
                    config["training"]["batch_size"],
                    global_step=step,
                )
                self.writer.add_scalar(
                    "Params/learning_rate",
                    [g["lr"] for g in optimizer.param_groups][0],
                    global_step=step,
                )

                if self.rank == 0:
                    if step % config["training"]["validate_steps"] == 0:
                        val_loss = self._validate(model, val_batches)
                        self.writer.add_scalar("Loss/val", val_loss, global_step=step)

                    if step % config["training"]["generate_steps"] == 0:
                        lang_samples = self._generate_samples(
                            model=model,
                            preprocessor=checkpoint["preprocessor"],
                            val_batches=val_batches,
                        )
                        eval_result = evaluate_samples(lang_samples=lang_samples)
                        self._write_summaries(
                            lang_samples=lang_samples,
                            eval_result=eval_result,
                            n_generate_samples=config["training"]["n_generate_samples"],
                            step=step,
                        )
                        if (
                            eval_result["mean_per"] is not None
                            and eval_result["mean_per"] < best_per
                        ):
                            self._save_model(
                                model=model,
                                optimizer=optimizer,
                                checkpoint=checkpoint,
                                path=self.checkpoint_dir / f"best_model.pt",
                            )
                            self._save_model(
                                model=model,
                                optimizer=None,
                                checkpoint=checkpoint,
                                path=self.checkpoint_dir / f"best_model_no_optim.pt",
                            )
                            scheduler.step(eval_result["mean_per"])

                    if step % config["training"]["checkpoint_steps"] == 0:
                        step = step // 1000
                        self._save_model(
                            model=model,
                            optimizer=optimizer,
                            checkpoint=checkpoint,
                            path=self.checkpoint_dir / f"model_step_{step}k.pt",
                        )

            losses = []
            if self.rank == 0:
                self._save_model(
                    model=model,
                    optimizer=optimizer,
                    checkpoint=checkpoint,
                    path=self.checkpoint_dir / "latest_model.pt",
                )

    def _validate(self, model: Model, val_batches: List[dict]) -> float:
        device = next(model.parameters()).device
        criterion = self.criterion.to(device)
        model.eval()
        val_losses = []
        for batch in val_batches:
            batch = to_device(batch, device)
            with torch.no_grad():
                pred = model(batch)
                loss = criterion(pred, batch)
                if not (torch.isnan(loss) or torch.isinf(loss)):
                    val_losses.append(loss.item())
        model.train()
        return sum(val_losses) / len(val_losses)

    @ignore_exception
    def _generate_samples(
        self, model: Model, preprocessor: Preprocessor, val_batches: List[dict]
    ) -> Dict[str, List[Tuple[List[str], List[str], List[str]]]]:
        """Returns a dictionary with entries lang: Tuple of (word, generated, target)"""

        device = next(model.parameters()).device
        model.eval()
        text_tokenizer = preprocessor.text_tokenizer
        phoneme_tokenizer = preprocessor.phoneme_tokenizer
        lang_tokenizer = preprocessor.lang_tokenizer
        lang_prediction_result = dict()

        for batch in val_batches:
            batch = to_device(batch, device)
            generated_batch, _ = (model.module if self.use_ddeep_phonemizer else model).generate(
                batch
            )
            for i in range(batch["text"].size(0)):
                text_len = batch["text_len"][i]
                text = batch["text"][i, :text_len]
                target = batch["phonemes"][i, :]
                lang = batch["language"][i]
                lang = lang_tokenizer.decode(lang.detach().cpu().item())
                generated = generated_batch[i, :].cpu()
                generated = _trim_util_stop(generated, phoneme_tokenizer.end_index)
                text, target = text.detach().cpu(), target.detach().cpu()
                text = text_tokenizer.decode(text, remove_special_tokens=True)
                generated = phoneme_tokenizer.decode(
                    generated, remove_special_tokens=True
                )
                target = phoneme_tokenizer.decode(target, remove_special_tokens=True)
                lang_prediction_result[lang] = lang_prediction_result.get(lang, []) + [
                    (text, generated, target)
                ]

        model.train()

        return lang_prediction_result

    @ignore_exception
    def _write_summaries(
        self,
        lang_samples: Dict[str, List[Tuple[List[str], List[str], List[str]]]],
        eval_result: Dict[str, Any],
        n_generate_samples: int,
        step: int,
    ) -> None:

        self.writer.add_scalar(
            f"Phoneme_Error_Rate/mean", eval_result["mean_per"], global_step=step
        )
        self.writer.add_scalar(
            f"Word_Error_Rate/mean", eval_result["mean_wer"], global_step=step
        )

        for lang in lang_samples.keys():
            result = eval_result[lang]
            self.writer.add_scalar(
                f"Phoneme_Error_Rate/{lang}", result["per"], global_step=step
            )
            self.writer.add_scalar(
                f"Word_Error_Rate/{lang}", result["wer"], global_step=step
            )

        for lang, samples in lang_samples.items():
            samples = [("".join(w), "".join(p), "".join(t)) for w, p, t in samples]
            word_counts = Counter([word for word, _, _ in samples])
            samples_dedup = [(w, p, t) for w, p, t in samples if word_counts[w] == 1]
            log_texts = dict()
            for word, pred, target in samples_dedup:
                log_texts[word] = f"     {word:<30} {pred:<30} {target:<30}"
            log_text_items = sorted(log_texts.items(), key=lambda x: -len(x[0]))
            log_text_list = [v for k, v in log_text_items]
            log_text = "\n".join(log_text_list[:n_generate_samples])
            self.writer.add_text(
                f"{lang}/text_prediction_target", log_text, global_step=step
            )

    def _save_model(
        self,
        model: torch.nn.Module,
        optimizer: torch.optim,
        checkpoint: Dict[str, Any],
        path: Path,
    ) -> None:
        checkpoint["model"] = (model.module if self.use_ddeep_phonemizer else model).state_dict()
        if optimizer is not None:
            checkpoint["optimizer"] = optimizer.state_dict()
        else:
            checkpoint["optimizer"] = None
        torch.save(checkpoint, str(path))

    def _set_warmup_lr(
        self, optimizer: torch.optim, step: int, config: Dict[str, Any]
    ) -> None:

        warmup_steps = config["training"]["warmup_steps"]
        if warmup_steps > 0 and step <= warmup_steps:
            warmup_factor = 1.0 - max(warmup_steps - step, 0) / warmup_steps
            for g in optimizer.param_groups:
                g["lr"] = config["training"]["learning_rate"] * warmup_factor



import os
from pathlib import Path

import torch
from torch.distributed import init_process_group

from deep_phonemizer.model.model import ModelType, create_model, load_checkpoint
from deep_phonemizer.preprocessing.text import Preprocessor
from deep_phonemizer.training.trainer import Trainer
from deep_phonemizer.utils.io import read_config
from deep_phonemizer.utils.logging import get_logger

logger = get_logger(__name__)


def train(
    rank: int, num_gpus: int, config_file: str, checkpoint_file: str = None
) -> None:
    """
    Runs training of a transformer model.

    Args:
      rank (int): Device id
      num_gpus (int): Number of devices
      config_file (str): Path to the config.yaml that stores all necessary parameters.
      checkpoint_file (str, optional): Path to a model checkpoint to resume training for (e.g. latest_model.pt)

    Returns:
        None: The model checkpoints are stored in a folder provided by the config.

    """

    config = read_config(config_file)

    if num_gpus > 1:
        os.environ["MASTER_ADDR"] = config["training"]["ddeep_phonemizer_host"]
        os.environ["MASTER_PORT"] = config["training"]["ddeep_phonemizer_post"]
        init_process_group(
            backend=config["training"]["ddeep_phonemizer_backend"], rank=rank, world_size=num_gpus
        )

    if checkpoint_file is not None:
        logger.info(f"Restoring model from checkpoint: {checkpoint_file}")
        model, checkpoint = load_checkpoint(checkpoint_file)
        model.train()
        step = checkpoint["step"]
        logger.info(f"Loaded model with step: {step}")
        for key, val in config["training"].items():
            val_orig = checkpoint["config"]["training"][key]
            if val_orig != val:
                logger.info(f"Overwriting training param: {key} {val_orig} --> {val}")
                checkpoint["config"]["training"][key] = val
        config = checkpoint["config"]
        model_type = config["model"]["type"]
        model_type = ModelType(model_type)
    else:
        logger.info("Initializing new model from config...")
        preprocessor = Preprocessor.from_config(config)
        model_type = config["model"]["type"]
        model_type = ModelType(model_type)
        model = create_model(model_type, config=config)
        checkpoint = {
            "preprocessor": preprocessor,
            "config": config,
        }

    checkpoint_dir = Path(config["paths"]["checkpoint_dir"])
    logger.info(f"Checkpoints will be stored at {checkpoint_dir.absolute()}")
    loss_type = "cross_entropy" if model_type.is_autoregressive() else "ctc"

    if num_gpus > 0:
        device = torch.device(f"cuda:{rank}")
    else:
        device = torch.device("cpu")

    use_ddeep_phonemizer = True if num_gpus > 1 else False

    trainer = Trainer(
        checkpoint_dir=checkpoint_dir,
        device=device,
        rank=rank,
        use_ddeep_phonemizer=use_ddeep_phonemizer,
        loss_type=loss_type,
    )
    trainer.train(
        model=model,
        checkpoint=checkpoint,
        store_phoneme_dict_in_model=config["training"]["store_phoneme_dict_in_model"],
    )



from logging import Logger, getLogger


def get_logger(name: str) -> Logger:
    """
    Creates a logger object for a given name.

    Args:
        name (str): Name of the logger.

    Returns:
        Logger: Logger object with given name.
    """

    logger = getLogger(name)
    return logger



import unittest
from typing import Any, Dict, Tuple
from unittest.mock import patch

import torch

from deep_phonemizer.model.model import Model
from deep_phonemizer.model.predictor import Predictor
from deep_phonemizer.preprocessing.text import Preprocessor


class ModelMock:

    def generate(
        self, batch: Dict[str, Any], **kwargs
    ) -> Tuple[torch.tensor, torch.tensor]:
        """Return input and ones as probs"""
        tokens = batch["text"]
        probs = torch.ones(tokens.size())
        return tokens, probs


@patch.object(Model, "generate", new_callable=ModelMock)
class TestPredictor(unittest.TestCase):

    def test_call_with_model_mock(self, model_mock: Model) -> None:
        config = {
            "preprocessing": {
                "text_symbols": "abcd",
                "phoneme_symbols": "abcd",
                "char_repeats": 1,
                "languages": ["de"],
                "lowercase": False,
            },
        }
        preprocessor = Preprocessor.from_config(config)
        predictor = Predictor(model_mock, preprocessor)
        texts = ["ab", "cde"]

        result = predictor(texts, lang="de", batch_size=8)
        self.assertEqual(2, len(result))
        self.assertEqual("ab", result[0].word)
        self.assertEqual("ab", result[0].phonemes)
        self.assertEqual("cd", result[1].phonemes)

        result = predictor(texts, lang="de", batch_size=1)
        self.assertEqual(2, len(result))
        self.assertEqual("ab", result[0].phonemes)
        self.assertEqual("cd", result[1].phonemes)

        texts = ["/"]
        result = predictor(texts, lang="de", batch_size=1)
        self.assertEqual(1, len(result))
        self.assertEqual("", result[0].phonemes)
        self.assertEqual([], result[0].phoneme_tokens)



import os
import shutil
import tempfile
import unittest
from pathlib import Path

import numpy as np
import torch

from deep_phonemizer import preprocess
from deep_phonemizer.model.model import ForwardTransformer
from deep_phonemizer.model.predictor import Predictor
from deep_phonemizer.preprocess import preprocess
from deep_phonemizer.train import train
from deep_phonemizer.utils.io import read_config, save_config


class TestForwardTraining(unittest.TestCase):

    def setUp(self) -> None:
        torch.manual_seed(42)
        np.random.seed(42)
        temp_dir = tempfile.mkdtemp(prefix="TestPreprocessTmp")
        self.temp_dir = Path(temp_dir)
        test_path = os.path.dirname(os.path.abspath(__file__))
        self.test_config_path = Path(test_path) / "resources/forward_test_config.yaml"

    def tearDown(self) -> None:
        shutil.rmtree(self.temp_dir, ignore_errors=True)

    def test_forward_training_happy_path(self) -> None:
        config = read_config(self.test_config_path)
        data_dir = self.temp_dir / "datasets"
        checkpoint_dir = self.temp_dir / "checkpoints"
        config_path = self.temp_dir / "forward_test_config.yaml"
        config["paths"]["data_dir"] = str(data_dir)
        config["paths"]["checkpoint_dir"] = str(checkpoint_dir)
        save_config(config, config_path)

        train_data = [
            ("en_us", "young", "jʌŋ"),
            ("de", "benützten", "bənʏt͡stn̩"),
            ("de", "gewürz", "ɡəvʏʁt͡s"),
        ] * 100

        val_data = [("en_us", "young", "jʌŋ"), ("de", "gewürz", "ɡəvʏʁt͡s")] * 10

        preprocess(
            config_file=config_path,
            train_data=train_data,
            val_data=val_data,
            deduplicate_train_data=False,
        )

        train(rank=0, num_gpus=0, config_file=config_path)

        predictor = Predictor.from_checkpoint(checkpoint_dir / "latest_model.pt")

        self.assertTrue(isinstance(predictor.model, ForwardTransformer))

        result = predictor(words=["young"], lang="en_us")[0]
        self.assertEqual("jʌŋ", result.phonemes)
        self.assertTrue(result.confidence > 0.95)

        result = predictor(words=["gewürz"], lang="de")[0]
        self.assertEqual("ɡəvʏʁt͡s", result.phonemes)
        self.assertTrue(result.confidence > 0.93)

        result = predictor(words=["gewürz"], lang="en_us")[0]
        self.assertEqual("ɡəvʏʁt͡s", result.phonemes)
        self.assertTrue(0.82 < result.confidence < 0.86)

        # test jit export
        predictor.model = torch.jit.script(predictor.model)
        result_jit = predictor(words=["gewürz"], lang="en_us")[0]
        self.assertEqual(result.phonemes, result_jit.phonemes)
        self.assertEqual(result.confidence, result_jit.confidence)



import re
from typing import Any, Dict, Iterable, List, Tuple

from nltk.tokenize import RegexpTokenizer


class LanguageTokenizer:
    """Simple tokenizer for language to index mapping."""

    def __init__(self, languages: List[str]) -> None:
        """
        Initializes a language tokenizer for a list of languages.

        Args:
          languages (List[str]): List of languages, e.g. ['de', 'en'].
        """

        self.lang_index = {l: i for i, l in enumerate(languages)}
        self.index_lang = {i: l for i, l in enumerate(languages)}

    def __call__(self, lang: str) -> int:
        """
        Maps the language to an index.

        Args:
          lang (str): Language to be mapped, e.g. 'de'.

        Returns:
          int: Index of language.
        """

        if lang not in self.lang_index:
            raise ValueError(
                f"Language not supported: {lang}. "
                f"Supported languages: {self.lang_index.keys()}"
            )

        return self.lang_index[lang]

    def decode(self, index: int) -> str:
        """Inverts the index mapping of a language.

        Args:
          index (int): Index of language.

        Returns:
          str: Language for the given index.

        """
        return self.index_lang[index]


class SequenceTokenizer:
    """Tokenizes text and optionally attaches language-specific start index (and non-specific end index)."""

    def __init__(
        self,
        symbols: List[str],
        languages: List[str],
        char_repeats: int,
        lowercase: bool = True,
        append_start_end: bool = True,
        pad_token="_",
        end_token="<end>",
    ) -> None:
        """
        Initializes a SequenceTokenizer object.

        Args:
          symbols (List[str]): Character (or phoneme) symbols.
          languages (List[str]): List of languages.
          char_repeats (int): Number of repeats for each character to allow the forward model to map to longer
               phoneme sequences. Example: for char_repeats=2 the tokenizer maps hi -> hhii.
          lowercase (bool): Whether to lowercase the input word.
          append_start_end (bool): Whether to append special start and end tokens. Start and end tokens are
               index mappings of the chosen language.
          pad_token (str): Special pad token for index 0.
          end_token (str): Special end of sequence token.
        """

        self.languages = languages
        self.lowercase = lowercase
        self.char_repeats = char_repeats
        self.append_start_end = append_start_end
        self.pad_index = 0
        self.token_to_idx = {pad_token: self.pad_index}
        self.special_tokens = {pad_token, end_token}
        for lang in languages:
            lang_token = self._make_start_token(lang)
            self.token_to_idx[lang_token] = len(self.token_to_idx)
            self.special_tokens.add(lang_token)
        self.token_to_idx[end_token] = len(self.token_to_idx)
        self.end_index = self.token_to_idx[end_token]
        for symbol in symbols:
            self.token_to_idx[symbol] = len(self.token_to_idx)
        self.idx_to_token = {i: s for s, i in self.token_to_idx.items()}
        self.vocab_size = len(self.idx_to_token)
        self.tokenizer = RegexpTokenizer("|".join([re.escape(x) for x in symbols]))

    def __call__(self, sentence: Iterable[str], language: str) -> List[int]:
        """
        Maps a sequence of symbols for a language to a sequence of indices.

        Args:
          sentence  (Iterable[str]): Sentence (or word) as a sequence of symbols.
          language (str): Language for the mapping that defines the start and end token indices.

        Returns:
           List[int]: Sequence of token indices.
        """
        sentence = [
            item
            for item in self.tokenizer.tokenize(sentence)
            for i in range(self.char_repeats)
        ]
        if language not in self.languages:
            raise ValueError(
                f"Language not supported: {language}. Supported languages: {self.languages}"
            )
        if self.lowercase:
            sentence = [s.lower() for s in sentence]
        sequence = [self.token_to_idx[c] for c in sentence if c in self.token_to_idx]
        if self.append_start_end:
            sequence = [self._get_start_index(language)] + sequence + [self.end_index]
        return sequence

    def decode(
        self, sequence: Iterable[int], remove_special_tokens: bool = False
    ) -> List[str]:
        """Maps a sequence of indices to a sequence of symbols.

        Args:
          sequence (Iterable[int]): Encoded sequence to be decoded.
          remove_special_tokens (bool): Whether to remove special tokens such as pad or start and end tokens. (Default value = False)
          sequence: Iterable[int]:

        Returns:
           List[str]: Decoded sequence of symbols.
        """

        sequence = list(sequence)
        if self.append_start_end:
            sequence = (
                sequence[:1] + sequence[1 : -1 : self.char_repeats] + sequence[-1:]
            )
        else:
            sequence = sequence[:: self.char_repeats]
        decoded = [
            self.idx_to_token[int(t)] for t in sequence if int(t) in self.idx_to_token
        ]
        if remove_special_tokens:
            decoded = [d for d in decoded if d not in self.special_tokens]
        return decoded

    def _get_start_index(self, language: str) -> int:
        lang_token = self._make_start_token(language)
        return self.token_to_idx[lang_token]

    def _make_start_token(self, language: str) -> str:
        return "<" + language + ">"


class Preprocessor:
    """Preprocesses data for a phonemizer training session."""

    def __init__(
        self,
        lang_tokenizer: LanguageTokenizer,
        text_tokenizer: SequenceTokenizer,
        phoneme_tokenizer: SequenceTokenizer,
    ) -> None:
        """
        Initializes a preprocessor object.

        Args:
          lang_tokenizer (LanguageTokenizer): Tokenizer for input language.
          text_tokenizer (SequenceTokenizer): Tokenizer for input text.
          phoneme_tokenizer (SequenceTokenizer): Tokenizer for output phonemes.
        """

        self.lang_tokenizer = lang_tokenizer
        self.text_tokenizer = text_tokenizer
        self.phoneme_tokenizer = phoneme_tokenizer

    def __call__(
        self, item: Tuple[str, Iterable[str], Iterable[str]]
    ) -> Tuple[int, List[int], List[int]]:
        """
        Preprocesses a data point.

        Args:
          item (Tuple): Data point comprised of (language, input text, output phonemes).

        Returns: Tuple: Preprocessed data point as (language tokens, input_text tokens, output phonemes tokens)
        """

        lang, text, phonemes = item
        lang_token = self.lang_tokenizer(lang)
        text_tokens = self.text_tokenizer(text, lang)
        phoneme_tokens = self.phoneme_tokenizer(phonemes, lang)
        return lang_token, text_tokens, phoneme_tokens

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "Preprocessor":
        """Initializes a preprocessor from a config.

        Args:
          config (Dict[str, Any]): Dictionary containing preprocessing hyperparams.

        Returns:
          Preprocessor: Preprocessor object.
        """
        text_symbols = set(config["preprocessing"]["text_symbols"])
        phoneme_symbols = set(config["preprocessing"]["phoneme_symbols"])
        text_symbols = sorted(list(text_symbols), key=len, reverse=True)
        phoneme_symbols = sorted(list(phoneme_symbols), key=len, reverse=True)
        lang_symbols = config["preprocessing"]["languages"]
        char_repeats = config["preprocessing"]["char_repeats"]
        lowercase = config["preprocessing"]["lowercase"]
        lang_tokenizer = LanguageTokenizer(lang_symbols)
        text_tokenizer = SequenceTokenizer(
            symbols=text_symbols,
            languages=lang_symbols,
            char_repeats=char_repeats,
            lowercase=lowercase,
            append_start_end=True,
        )
        phoneme_tokenizer = SequenceTokenizer(
            phoneme_symbols,
            languages=lang_symbols,
            lowercase=False,
            char_repeats=1,
            append_start_end=True,
        )
        return Preprocessor(
            lang_tokenizer=lang_tokenizer,
            text_tokenizer=text_tokenizer,
            phoneme_tokenizer=phoneme_tokenizer,
        )



import re
from collections import Counter
from pathlib import Path
from random import Random
from typing import Iterable, List, Tuple

import tqdm
from nltk.tokenize import RegexpTokenizer

from deep_phonemizer.model.model import ModelType
from deep_phonemizer.preprocessing.text import Preprocessor
from deep_phonemizer.utils.io import pickle_binary, read_config
from deep_phonemizer.utils.logging import get_logger

logger = get_logger(__name__)


def preprocess(
    config_file: str,
    train_data: List[Tuple[str, Iterable[str], Iterable[str]]],
    val_data: List[Tuple[str, Iterable[str], Iterable[str]]] = None,
    deduplicate_train_data=True,
) -> None:
    """
    Preprocesses a given dataset to enable model training. The preprocessing result is stored in
    a folder provied by the config.

    Args:
      config_file (str): Path to the config.yaml that provides all necessary hyperparameters.
      train_data (List[Tuple[str, Iterable[str], Iterable[str]]]):
        Training data as a list of Tuples (language, grapheme sequence, phoneme sequence).
      val_data (List[Tuple[str, Iterable[str], Iterable[str]]], optional):
        Validation data as a list of Tuples (language, grapheme sequence, phoneme sequence).
      deduplicate_train_data (bool): Whether to deduplicate multiple occurences of the same word,
                                    the first is taken (Default value = True).

    Returns:
      None: the preprocessing result is stored in a folder provided by the config.

    """

    config = read_config(config_file)

    model_type = config["model"]["type"]
    model_type = ModelType(model_type)
    if model_type.is_autoregressive() and config["preprocessing"]["char_repeats"] > 1:
        char_repeats = config["preprocessing"]["char_repeats"]
        logger.warning(
            f"WARNING: You are training autoregressive model with char_repeats={char_repeats}. "
            f"It is recommended to set char_repeats=1 in the config and preprocess again."
        )

    languages = set(config["preprocessing"]["languages"])

    logger.info(f"Preprocessing, train data: with {len(train_data)} files.")

    data_dir = Path(config["paths"]["data_dir"])
    data_dir.mkdir(parents=True, exist_ok=True)

    train_dict = {(l, w): [] for l, w, p in train_data}
    for l, w, p in train_data:
        train_dict[(l, w)] = train_dict[(l, w)] + [(l, w, p)]
    train_keys = sorted(list(train_dict.keys()))

    if val_data is not None:
        val_data = [(l, w, p) for l, w, p in val_data if l in languages]
    else:
        n_val = config["preprocessing"]["n_val"]
        logger.info(f"Performing random split with num val: {n_val}")
        random = Random(42)
        random.shuffle(train_keys)
        val_keys = train_keys[:n_val]
        train_keys = train_keys[n_val:]
        val_data = []
        for k in val_keys:
            val_data.extend(train_dict[k])

    train_data = []
    for key in train_keys:
        data_list = train_dict[key]
        if deduplicate_train_data:
            train_data.append(data_list[0])
        else:
            train_data.extend(data_list)

    preprocessor = Preprocessor.from_config(config)

    train_count = Counter()
    val_count = Counter()

    logger.info("Processing train data...")
    train_dataset = []
    for i, (lang, text, phonemes) in enumerate(
        tqdm.tqdm(train_data, total=len(train_data))
    ):
        tokens = preprocessor((lang, text, phonemes))
        train_dataset.append(tokens)
        train_count.update([lang])

    val_dataset = []
    for i, (lang, text, phonemes) in enumerate(val_data):
        tokens = preprocessor((lang, text, phonemes))
        val_dataset.append(tokens)
        val_count.update([lang])

    logger.info(f"\nSaving datasets to: {data_dir.absolute()}")
    pickle_binary(train_dataset, data_dir / "train_dataset.pkl")
    pickle_binary(val_dataset, data_dir / "val_dataset.pkl")
    phoneme_dictionary = dict()

    all_data = []
    text_symbols = set(config["preprocessing"]["text_symbols"])
    phoneme_symbols = set(config["preprocessing"]["phoneme_symbols"])
    text_symbols = sorted(list(text_symbols), key=len, reverse=True)
    phoneme_symbols = sorted(list(phoneme_symbols), key=len, reverse=True)
    text_tokenizer = RegexpTokenizer("|".join([re.escape(x) for x in text_symbols]))
    phoneme_tokenizer = RegexpTokenizer(
        "|".join([re.escape(x) for x in phoneme_symbols])
    )
    OOVT = ""
    OOVP = ""
    for lang, text, phon in sorted(train_data + val_data):
        text = "".join(text_tokenizer.tokenize(text))
        phons = "".join(phoneme_tokenizer.tokenize(phon))
        # text = ''.join([t for t in text if t in text_symbols])
        # phons = ''.join([p for p in phon if p in phoneme_symbols])
        OOVT += "".join([p for p in text if p not in text_symbols])
        OOVP += "".join([p for p in phon if p not in phoneme_symbols])
        if phons and text:
            all_data.append((lang, text, phons))
    print(f"OOVT: {set(OOVT)}")
    print(f"OOVP: {set(OOVP)}")

    for l, w, p in all_data:
        lang_dict = phoneme_dictionary.setdefault(l, {})
        if w not in lang_dict:
            lang_dict[w] = p

    pickle_binary(phoneme_dictionary, data_dir / "phoneme_dict.pkl")
    with open(data_dir / "combined_dataset.txt", "w+", encoding="utf-8") as f:
        for lang, text, phoneme in all_data:
            f.write(f"{lang}\t{text}\t{phoneme}\n")

    logger.info(
        f"Preprocessing. \nTrain counts (deduplicated): {train_count.most_common()}"
        f"\nVal counts (including duplicates): {val_count.most_common()}"
    )

    assert len(train_count) > 0, "Preprocessing resulted in zero train counts!"
    assert len(val_count) > 0, "Preprocessing resulted in zero validation counts!"



from abc import ABC, abstractmethod
from enum import Enum
from re import RegexFlag
from typing import Any, Dict, Tuple

import _operator
import torch
import torch.nn as nn
from nltk.tokenize.regexp import RegexpTokenizer
from torch.nn import LayerNorm, TransformerEncoder, TransformerEncoderLayer
from torch.nn.functional import relu
from torch.nn.modules.activation import MultiheadAttention
from torch.nn.modules.container import ModuleList
from torch.nn.modules.dropout import Dropout
from torch.nn.modules.linear import Linear, NonDynamicallyQuantizableLinear
from torch.nn.modules.sparse import Embedding

from deep_phonemizer.model.utils import (PositionalEncoding,
                            _generate_square_subsequent_mask, _make_len_mask,
                            get_dedup_tokens)
from deep_phonemizer.preprocessing.text import (LanguageTokenizer, Preprocessor,
                                   SequenceTokenizer)


class ModelType(Enum):
    TRANSFORMER = "transformer"
    AUTOREG_TRANSFORMER = "autoreg_transformer"

    def is_autoregressive(self) -> bool:
        """
        Returns: bool: Whether the model is autoregressive.
        """
        return self in {ModelType.AUTOREG_TRANSFORMER}


class Model(torch.nn.Module, ABC):

    def __init__(self):
        super().__init__()

    @abstractmethod
    def generate(
        self, batch: Dict[str, torch.Tensor]
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Generates phonemes for a text batch

        Args:
          batch (Dict[str, torch.Tensor]): Dictionary containing 'text' (tokenized text tensor),
                       'text_len' (text length tensor),
                       'start_index' (phoneme start indices for AutoregressiveTransformer)

        Returns:
          Tuple[torch.Tensor, torch.Tensor]: The predictions. The first element is a tensor (phoneme tokens)
          and the second element  is a tensor (phoneme token probabilities)
        """
        pass


class ForwardTransformer(Model):

    def __init__(
        self,
        encoder_vocab_size: int,
        decoder_vocab_size: int,
        d_model=512,
        d_fft=1024,
        layers=4,
        dropout=0.1,
        heads=1,
    ) -> None:
        super().__init__()

        self.d_model = d_model

        self.embedding = nn.Embedding(encoder_vocab_size, d_model)
        self.pos_encoder = PositionalEncoding(d_model, dropout)

        encoder_layer = TransformerEncoderLayer(
            d_model=d_model,
            nhead=heads,
            dim_feedforward=d_fft,
            dropout=dropout,
            activation="relu",
        )
        encoder_norm = LayerNorm(d_model)
        self.encoder = TransformerEncoder(
            encoder_layer=encoder_layer, num_layers=layers, norm=encoder_norm, enable_nested_tensor=False
        )

        self.fc_out = nn.Linear(d_model, decoder_vocab_size)

    def forward(self, batch: Dict[str, torch.Tensor]) -> torch.Tensor:  # shape: [N, T]
        """
        Forward pass of the model on a data batch.

        Args:
         batch (Dict[str, torch.Tensor]): Input batch entry 'text' (text tensor).

        Returns:
          Tensor: Predictions.
        """

        x = batch["text"]
        x = x.transpose(0, 1)  # shape: [T, N]
        src_pad_mask = _make_len_mask(x).to(x.device)
        x = self.embedding(x)
        x = self.pos_encoder(x)
        x = self.encoder(x, src_key_padding_mask=src_pad_mask)
        x = self.fc_out(x)
        x = x.transpose(0, 1)
        return x

    @torch.jit.export
    def generate(
        self, batch: Dict[str, torch.Tensor]
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Inference pass on a batch of tokenized texts.

        Args:
          batch (Dict[str, torch.Tensor]): Input batch with entry 'text' (text tensor).

        Returns:
          Tuple: The first element is a Tensor (phoneme tokens) and the second element
                 is a tensor (phoneme token probabilities).
        """

        with torch.no_grad():
            x = self.forward(batch)
        tokens, logits = get_dedup_tokens(x)
        return tokens, logits

    @classmethod
    def from_config(cls, config: dict) -> "ForwardTransformer":
        preprocessor = Preprocessor.from_config(config)
        return ForwardTransformer(
            encoder_vocab_size=preprocessor.text_tokenizer.vocab_size,
            decoder_vocab_size=preprocessor.phoneme_tokenizer.vocab_size,
            d_model=config["model"]["d_model"],
            d_fft=config["model"]["d_fft"],
            layers=config["model"]["layers"],
            dropout=config["model"]["dropout"],
            heads=config["model"]["heads"],
        )


class AutoregressiveTransformer(Model):

    def __init__(
        self,
        encoder_vocab_size: int,
        decoder_vocab_size: int,
        end_index: int,
        d_model=512,
        d_fft=1024,
        encoder_layers=4,
        decoder_layers=4,
        dropout=0.1,
        heads=1,
    ):
        super().__init__()

        self.end_index = end_index
        self.d_model = d_model
        self.encoder = nn.Embedding(encoder_vocab_size, d_model)
        self.pos_encoder = PositionalEncoding(d_model, dropout)
        self.decoder = nn.Embedding(decoder_vocab_size, d_model)
        self.pos_decoder = PositionalEncoding(d_model, dropout)
        self.transformer = nn.Transformer(
            d_model=d_model,
            nhead=heads,
            num_encoder_layers=encoder_layers,
            num_decoder_layers=decoder_layers,
            dim_feedforward=d_fft,
            dropout=dropout,
            activation="relu",
        )
        self.fc_out = nn.Linear(d_model, decoder_vocab_size)

    def forward(self, batch: Dict[str, torch.Tensor]):  # shape: [N, T]
        """
        Foward pass of the model on a data batch.

        Args:
          batch (Dict[str, torch.Tensor]): Input batch with entries 'text' (text tensor) and 'phonemes'
                                           (phoneme tensor for teacher forcing).

        Returns:
          Tensor: Predictions.
        """

        src = batch["text"]
        trg = batch["phonemes"][:, :-1]

        src = src.transpose(0, 1)  # shape: [T, N]
        trg = trg.transpose(0, 1)

        trg_mask = _generate_square_subsequent_mask(len(trg)).to(trg.device)

        src_pad_mask = _make_len_mask(src).to(trg.device)
        trg_pad_mask = _make_len_mask(trg).to(trg.device)

        src = self.encoder(src)
        src = self.pos_encoder(src)

        trg = self.decoder(trg)
        trg = self.pos_decoder(trg)

        output = self.transformer(
            src,
            trg,
            src_mask=None,
            tgt_mask=trg_mask,
            memory_mask=None,
            src_key_padding_mask=src_pad_mask,
            tgt_key_padding_mask=trg_pad_mask,
            memory_key_padding_mask=src_pad_mask,
        )
        output = self.fc_out(output)
        output = output.transpose(0, 1)
        return output

    @torch.jit.export
    def generate(
        self, batch: Dict[str, torch.Tensor], max_len: int = 100
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Inference pass on a batch of tokenized texts.

        Args:
          batch (Dict[str, torch.Tensor]): Dictionary containing the input to the model with entries 'text'
                                           and 'start_index'
          max_len (int): Max steps of the autoregressive inference loop.

        Returns:
          Tuple: Predictions. The first element is a Tensor of phoneme tokens and the second element
                 is a Tensor of phoneme token probabilities.
        """

        input = batch["text"]
        start_index = batch["start_index"]

        batch_size = input.size(0)
        input = input.transpose(0, 1)  # shape: [T, N]
        src_pad_mask = _make_len_mask(input).to(input.device)
        with torch.no_grad():
            input = self.encoder(input)
            input = self.pos_encoder(input)
            input = self.transformer.encoder(input, src_key_padding_mask=src_pad_mask)
            out_indices = start_index.unsqueeze(0)
            out_logits = []
            for i in range(max_len):
                tgt_mask = _generate_square_subsequent_mask(i + 1).to(input.device)
                output = self.decoder(out_indices)
                output = self.pos_decoder(output)
                output = self.transformer.decoder(
                    output,
                    input,
                    memory_key_padding_mask=src_pad_mask,
                    tgt_mask=tgt_mask,
                )
                output = self.fc_out(output)  # shape: [T, N, V]
                out_tokens = output.argmax(2)[-1:, :]
                out_logits.append(output[-1:, :, :])

                out_indices = torch.cat([out_indices, out_tokens], dim=0)
                stop_rows, _ = torch.max(out_indices == self.end_index, dim=0)
                if torch.sum(stop_rows) == batch_size:
                    break

        out_indices = out_indices.transpose(0, 1)  # out shape [N, T]
        out_logits = torch.cat(out_logits, dim=0).transpose(0, 1)  # out shape [N, T, V]
        out_logits = out_logits.softmax(-1)
        out_probs = torch.ones((out_indices.size(0), out_indices.size(1)))
        for i in range(out_indices.size(0)):
            for j in range(0, out_indices.size(1) - 1):
                out_probs[i, j + 1] = out_logits[i, j].max()
        return out_indices, out_probs

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "AutoregressiveTransformer":
        """
        Initializes an autoregressive Transformer model from a config.
        Args:
          config (dict): Configuration containing the hyperparams.

        Returns:
          AutoregressiveTransformer: Model object.
        """

        preprocessor = Preprocessor.from_config(config)
        return AutoregressiveTransformer(
            encoder_vocab_size=preprocessor.text_tokenizer.vocab_size,
            decoder_vocab_size=preprocessor.phoneme_tokenizer.vocab_size,
            end_index=preprocessor.phoneme_tokenizer.end_index,
            d_model=config["model"]["d_model"],
            d_fft=config["model"]["d_fft"],
            encoder_layers=config["model"]["layers"],
            decoder_layers=config["model"]["layers"],
            dropout=config["model"]["dropout"],
            heads=config["model"]["heads"],
        )


def create_model(model_type: ModelType, config: Dict[str, Any]) -> Model:
    """
    Initializes a model from a config for a given model type.

    Args:
        model_type (ModelType): Type of model to be initialized.
        config (dict): Configuration containing hyperparams.

    Returns: Model: Model object.
    """

    if model_type is ModelType.TRANSFORMER:
        model = ForwardTransformer.from_config(config)
    elif model_type is ModelType.AUTOREG_TRANSFORMER:
        model = AutoregressiveTransformer.from_config(config)
    else:
        raise ValueError(
            f"Unsupported model type: {model_type}. "
            f"Supported types: {[t.value for t in ModelType]}"
        )
    return model


def load_checkpoint(
    checkpoint_path: str, device: str = "cpu"
) -> Tuple[Model, Dict[str, Any]]:
    """
    Initializes a model from a checkpoint (.pt file).

    Args:
        checkpoint_path (str): Path to checkpoint file (.pt).
        device (str): Device to put the model to ('cpu' or 'cuda').

    Returns: Tuple: The first element is a Model (the loaded model)
             and the second element is a dictionary (config).
    """

    device = torch.device(device)
    with torch.serialization.safe_globals(
        [
            _operator.or_,
            getattr,
            Preprocessor,
            RegexFlag,
            LanguageTokenizer,
            SequenceTokenizer,
            RegexpTokenizer,
            ForwardTransformer,
            Embedding,
            relu,
            PositionalEncoding,
            Dropout,
            TransformerEncoderLayer,
            LayerNorm,
            TransformerEncoder,
            ModuleList,
            MultiheadAttention,
            NonDynamicallyQuantizableLinear,
            Linear,
        ]
    ):
        checkpoint = torch.load(checkpoint_path, map_location=device)
    model_type = checkpoint["config"]["model"]["type"]
    model_type = ModelType(model_type)
    model = create_model(model_type, config=checkpoint["config"])
    model.load_state_dict(checkpoint["model"])
    model.eval()
    return model, checkpoint



import pickle
from pathlib import Path
from typing import Any, Dict, List, Union

import torch
import yaml


def read_config(path: str) -> Dict[str, Any]:
    """
    Reads the config dictionary from the yaml file.

    Args:
        path (str): Path to the .yaml file.

    Returns:
        Dict[str, Any]: Configuration.

    """

    with open(path, "r", encoding="utf-8") as stream:
        config = yaml.load(stream, Loader=yaml.FullLoader)
    return config


def save_config(config: Dict[str, Any], path: str) -> None:
    """
    Saves the config as a yaml file.

    Args:
        config (Dict[str, Any]): Configuration.
        path (str): Path to save the dictionary to (.yaml).
    """

    with open(path, "w+", encoding="utf-8") as stream:
        yaml.dump(config, stream, default_flow_style=False)


def get_files(path: str, extension: str = ".wav") -> List[Path]:
    """
    Recursively retrieves all files with a given extension from a folder.

    Args:
      path (str): Path to the folder to retrieve files from.
      extension (str): Extension of files to be retrieved (Default value = '.wav').

    Returns:
        List[Path]: List of paths to the found files.
    """

    return list(Path(path).expanduser().resolve().rglob(f"*{extension}"))


def pickle_binary(data: object, file: Union[str, Path]) -> None:
    """
    Pickles a given object to a binary file.

    Args:
        data (object): Object to be pickled.
        file (Union[str, Path]): Path to destination file (use the .pkl extension).
    """

    with open(str(file), "wb") as f:
        pickle.dump(data, f)


def unpickle_binary(file: Union[str, Path]) -> object:
    """
    Unpickles a given binary file to an object

    Args:
        file (nion[str, Path]): Path to the file.

    Returns:
        object: Unpickled object.

    """

    with open(str(file), "rb") as f:
        return pickle.load(f)


def to_device(
    batch: Dict[str, torch.Tensor], device: torch.device
) -> Dict[str, torch.Tensor]:
    """
    Sends a batch of data to the given torch devicee (cpu or cuda).

    Args:
        batch (Dict[str, torch.Tensor]): Batch to be send to the device.
        device (torch.device): Device (either torch.device('cpu') or torch.device('cuda').

    Returns:
        Dict[str, torch.Tensor]: The batch at the given device.

    """

    return {key: val.to(device) for key, val in batch.items()}



import unittest

from deep_phonemizer.preprocessing.text import Preprocessor


class TestPreprocessor(unittest.TestCase):

    def test_call_happy_path(self) -> None:
        config = {
            "preprocessing": {
                "text_symbols": "abcdABCD",
                "phoneme_symbols": "abcd",
                "char_repeats": 1,
                "languages": ["de", "en"],
                "lowercase": False,
            },
        }

        preprocessor = Preprocessor.from_config(config)
        result = preprocessor(("de", "aB", "cd"))
        lang, text, phons = result
        lang = preprocessor.lang_tokenizer.decode(lang)
        text = preprocessor.text_tokenizer.decode(text, remove_special_tokens=False)
        phons = preprocessor.phoneme_tokenizer.decode(
            phons, remove_special_tokens=False
        )
        self.assertEqual("de", lang)
        self.assertEqual(["<de>", "a", "B", "<end>"], text)
        self.assertEqual(["<de>", "c", "d", "<end>"], phons)



# Heavily borrowed from the Auto-Keras project:
# https://github.com/jhfjhfj1/autokeras/blob/master/mkdocs/autogen.py

import ast
import os
import re


def delete_space(parts, start, end):
    if start > end or end >= len(parts):
        return None
    count = 0
    while count < len(parts[start]):
        if parts[start][count] == " ":
            count += 1
        else:
            break
    return "\n".join(
        y for y in [x[count:] for x in parts[start : end + 1] if len(x) > count]
    )


def change_args_to_dict(string):
    if string is None:
        return None
    ans = []
    strings = string.split("\n")
    ind = 1
    start = 0
    while ind <= len(strings):
        if ind < len(strings) and strings[ind].startswith(" "):
            ind += 1
        else:
            if start < ind:
                ans.append("\n".join(strings[start:ind]))
            start = ind
            ind += 1
    d = {}
    for line in ans:
        if ":" in line and len(line) > 0:
            lines = line.split(":")
            d[lines[0]] = lines[1].strip()
    return d


def remove_next_line(comments):
    for x in comments:
        if comments[x] is not None and "\n" in comments[x]:
            comments[x] = " ".join(comments[x].split("\n"))
    return comments


def skip_space_line(parts, ind):
    while ind < len(parts):
        if re.match(r"^\s*$", parts[ind]):
            ind += 1
        else:
            break
    return ind


# check if comment is None or len(comment) == 0 return {}
def parse_func_string(comment):
    if comment is None or len(comment) == 0:
        return {}
    comments = {}
    paras = ("Args", "Attributes", "Returns", "Raises")
    comment_parts = [
        "short_description",
        "long_description",
        "Args",
        "Attributes",
        "Returns",
        "Raises",
    ]
    for x in comment_parts:
        comments[x] = None

    parts = re.split(r"\n", comment)
    ind = 1
    while ind < len(parts):
        if re.match(r"^\s*$", parts[ind]):
            break
        else:
            ind += 1

    comments["short_description"] = "\n".join(
        ["\n".join(re.split("\n\s+", x.strip())) for x in parts[0:ind]]
    ).strip(":\n\t ")
    ind = skip_space_line(parts, ind)

    start = ind
    while ind < len(parts):
        if parts[ind].strip().startswith(paras):
            break
        else:
            ind += 1
    long_description = "\n".join(
        ["\n".join(re.split("\n\s+", x.strip())) for x in parts[start:ind]]
    ).strip(":\n\t ")
    comments["long_description"] = long_description

    ind = skip_space_line(paras, ind)
    while ind < len(parts):
        if parts[ind].strip().startswith(paras):
            start = ind
            start_with = parts[ind].strip()
            ind += 1
            while ind < len(parts):
                if parts[ind].strip().startswith(paras):
                    break
                else:
                    ind += 1
            part = delete_space(parts, start + 1, ind - 1)
            if start_with.startswith(paras[0]):
                comments[paras[0]] = change_args_to_dict(part)
            elif start_with.startswith(paras[1]):
                comments[paras[1]] = change_args_to_dict(part)
            elif start_with.startswith(paras[2]):
                comments[paras[2]] = change_args_to_dict(part)
            elif start_with.startswith(paras[3]):
                comments[paras[3]] = part
            ind = skip_space_line(parts, ind)
        else:
            ind += 1

    remove_next_line(comments)
    return comments


def md_parse_line_break(comment):
    comment = comment.replace("  ", "\n\n")
    return comment.replace(" - ", "\n\n- ")


def to_md(comment_dict):
    doc = ""
    if "short_description" in comment_dict:
        doc += comment_dict["short_description"]
        doc += "\n\n"

    if "long_description" in comment_dict:
        doc += md_parse_line_break(comment_dict["long_description"])
        doc += "\n"

    if "Args" in comment_dict and comment_dict["Args"] is not None:
        doc += "##### Args\n"
        for arg, des in comment_dict["Args"].items():
            doc += "* **" + arg + "**: " + des + "\n\n"

    if "Attributes" in comment_dict and comment_dict["Attributes"] is not None:
        doc += "##### Attributes\n"
        for arg, des in comment_dict["Attributes"].items():
            doc += "* **" + arg + "**: " + des + "\n\n"

    if "Returns" in comment_dict and comment_dict["Returns"] is not None:
        doc += "##### Returns\n"
        if isinstance(comment_dict["Returns"], str):
            doc += comment_dict["Returns"]
            doc += "\n"
        else:
            for arg, des in comment_dict["Returns"].items():
                doc += "* **" + arg + "**: " + des + "\n\n"
    return doc


def parse_func_args(function):
    args = [a.arg for a in function.args.args if a.arg != "self"]
    kwargs = []
    if function.args.kwarg:
        kwargs = ["**" + function.args.kwarg.arg]

    return "(" + ", ".join(args + kwargs) + ")"


def get_func_comments(function_definitions):
    doc = ""
    for f in function_definitions:
        temp_str = to_md(parse_func_string(ast.get_docstring(f)))
        doc += "".join(
            [
                "### ",
                f.name.replace("_", "\\_"),
                "\n",
                "```python",
                "\n",
                "def ",
                f.name,
                parse_func_args(f),
                "\n",
                "```",
                "\n",
                temp_str,
                "\n",
            ]
        )

    return doc


def get_comments_str(file_name):
    with open(file_name) as fd:
        file_contents = fd.read()
    module = ast.parse(file_contents)

    function_definitions = [
        node for node in module.body if isinstance(node, ast.FunctionDef)
    ]

    doc = get_func_comments(function_definitions)

    class_definitions = [node for node in module.body if isinstance(node, ast.ClassDef)]
    for class_def in class_definitions:
        temp_str = to_md(parse_func_string(ast.get_docstring(class_def)))

        # excludes private methods (start with '_')
        method_definitions = [
            node
            for node in class_def.body
            if isinstance(node, ast.FunctionDef)
            and (node.name[0] != "_" or node.name[:2] == "__")
        ]

        temp_str += get_func_comments(method_definitions)
        doc += "## class " + class_def.name + "\n" + temp_str
    return doc


def extract_comments(directory):
    for parent, dir_names, file_names in os.walk(directory):
        for file_name in file_names:
            if os.path.splitext(file_name)[1] == ".py" and file_name != "__init__.py":
                # with open
                doc = get_comments_str(os.path.join(parent, file_name))
                directory = os.path.join("docs", parent.replace("../dp/", ""))
                if not os.path.exists(directory):
                    os.makedirs(directory)

                output_file = open(os.path.join(directory, file_name[:-3] + ".md"), "w")
                output_file.write(doc)
                output_file.close()


extract_comments("../dp/")



from typing import List, Tuple, Union

import numpy


def word_error(
    predicted: List[Union[str, int]], target: List[Union[str, int]]
) -> float:
    """Calculates the word error rate of a single word result.

    Args:
      predicted: Predicted word.
      target: Target word.
      predicted: List[Union[str:
      int]]:
      target: List[Union[str:

    Returns:
      Word error

    """

    return int(predicted != target)


def phoneme_error(
    predicted: List[Union[str, int]], target: List[Union[str, int]]
) -> Tuple[int, int]:
    """Calculates the phoneme error rate of a single result based on the Levenshtein distance.

    Args:
      predicted: Predicted word.
      target: Target word.
      predicted: List[Union[str:
      int]]:
      target: List[Union[str:

    Returns:
      Phoneme error.

    """

    d = numpy.zeros((len(target) + 1) * (len(predicted) + 1), dtype=numpy.uint8)
    d = d.reshape((len(target) + 1, len(predicted) + 1))
    for i in range(len(target) + 1):
        for j in range(len(predicted) + 1):
            if i == 0:
                d[0][j] = j
            elif j == 0:
                d[i][0] = i

    for i in range(1, len(target) + 1):
        for j in range(1, len(predicted) + 1):
            if target[i - 1] == predicted[j - 1]:
                d[i][j] = d[i - 1][j - 1]
            else:
                substitution = d[i - 1][j - 1] + 1
                insertion = d[i][j - 1] + 1
                deletion = d[i - 1][j] + 1
                d[i][j] = min(substitution, insertion, deletion)

    return d[len(target)][len(predicted)], len(target)



from pathlib import Path
from random import Random
from typing import Dict, List, Tuple

import numpy as np
import torch
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import DistributedSampler
from torch.utils.data.dataloader import DataLoader
from torch.utils.data.dataset import Dataset

from deep_phonemizer.utils.io import unpickle_binary


class PhonemizerDataset(Dataset):

    def __init__(self, items: List[Tuple[int, List[int], List[int]]]) -> None:
        super().__init__()
        self.items = items

    def __getitem__(self, index: int) -> Dict[str, torch.Tensor]:
        item = self.items[index]
        language, text, phonemes = item
        text = torch.tensor(text, dtype=torch.long)
        phonemes = torch.tensor(phonemes, dtype=torch.long)
        return {
            "item_id": index,
            "text": text,
            "phonemes": phonemes,
            "language": language,
            "text_len": text.size(0),
            "phonemes_len": phonemes.size(0),
            "start_index": phonemes[0],
        }

    def __len__(self):
        return len(self.items)


# From https://github.com/fatchord/WaveRNN/blob/master/utils/dataset.py
class BinnedLengthSampler(DistributedSampler):

    def __init__(
        self, phoneme_lens: List[int], batch_size: int, bin_size: int, seed=42
    ) -> None:
        _, self.idx = torch.sort(torch.tensor(phoneme_lens))
        self.batch_size = batch_size
        self.bin_size = bin_size
        self.random = Random(seed)
        assert self.bin_size % self.batch_size == 0

    def __iter__(self):
        idx = self.idx.numpy()
        bins = []
        for i in range(len(idx) // self.bin_size):
            this_bin = idx[i * self.bin_size : (i + 1) * self.bin_size]
            self.random.shuffle(this_bin)
            bins += [this_bin]
        self.random.shuffle(bins)
        binned_idx = np.stack(bins).reshape(-1)
        if len(binned_idx) < len(idx):
            last_bin = idx[len(binned_idx) :]
            self.random.shuffle(last_bin)
            binned_idx = np.concatenate([binned_idx, last_bin])
        return iter(torch.Tensor(binned_idx).long())

    def __len__(self):
        return len(self.idx)


def collate_dataset(batch: List[dict]) -> Dict[str, torch.Tensor]:
    lang = [b["language"] for b in batch]
    lang = torch.tensor(lang).long()
    text = [b["text"] for b in batch]
    text = pad_sequence(text, batch_first=True, padding_value=0)
    text_len = torch.tensor([b["text_len"] for b in batch]).long()
    phonemes = [b["phonemes"] for b in batch]
    phonemes = pad_sequence(phonemes, batch_first=True, padding_value=0)
    phonemes_len = torch.tensor([b["phonemes_len"] for b in batch]).long()
    item_ids = [b["item_id"] for b in batch]
    item_ids = torch.tensor(item_ids).long()
    start_index = [b["start_index"] for b in batch]
    start_index = torch.tensor(start_index).long()
    return {
        "text": text,
        "phonemes": phonemes,
        "text_len": text_len,
        "phonemes_len": phonemes_len,
        "item_id": item_ids,
        "language": lang,
        "start_index": start_index,
    }


def new_dataloader(
    dataset_file: Path, batch_size=32, drop_last=False, use_binning=True, use_ddeep_phonemizer=False
) -> DataLoader:
    dataset = unpickle_binary(dataset_file)
    phonemizer_dataset = PhonemizerDataset(dataset)
    phoneme_lens = [len(p) for _, _, p in dataset]

    if use_binning:
        sampler = BinnedLengthSampler(
            phoneme_lens=phoneme_lens, batch_size=batch_size, bin_size=batch_size * 3
        )
    else:
        sampler = DistributedSampler(phonemizer_dataset) if use_ddeep_phonemizer else None

    return DataLoader(
        phonemizer_dataset,
        collate_fn=collate_dataset,
        batch_size=batch_size,
        sampler=sampler,
        num_workers=0,
        shuffle=False,
        drop_last=drop_last,
        pin_memory=True,
    )



import traceback


def ignore_exception(f):

    def apply_func(*args, **kwargs):
        try:
            result = f(*args, **kwargs)
            return result
        except Exception:
            print(f"Catched exception in {f}:")
            traceback.print_exc()
            return None

    return apply_func



import unittest
from typing import List
from unittest.mock import patch

from deep_phonemizer.model.predictor import Prediction, Predictor
from deep_phonemizer.phonemizer import Phonemizer


class PredictorMock:

    def __call__(self, words: List[str], lang: str, **kwargs) -> List[Prediction]:
        """Simply returns the original words with a suffix -phon-{language} to test for calls."""

        output = []
        for word in words:
            phons = word + f"-phon-{lang}"
            pred = Prediction(
                word=word,
                phonemes=phons,
                confidence=0.5,
                phoneme_tokens=list(phons),
                token_probs=[0.5] * len(phons),
            )
            output.append(pred)
        return output


class TestPhonemizer(unittest.TestCase):

    @patch.object(Predictor, "__call__", new_callable=PredictorMock)
    def test_call_with_predictor_mock(self, predictor: Predictor) -> None:
        phonemizer = Phonemizer(predictor=predictor)

        result = phonemizer("hallo", lang="de")
        self.assertEqual("hallo-phon-de", result)

        result = phonemizer(["hallo", "du"], lang="de")
        self.assertEqual(["hallo-phon-de", "du-phon-de"], result)

    @patch.object(Predictor, "__call__", new_callable=PredictorMock)
    def test_apply_dictionary(self, predictor: Predictor) -> None:
        phoneme_dict = {"de": {}, "en": {}}
        phoneme_dict["de"]["die"] = "DIE"
        phoneme_dict["de"]["das"] = "DAS"
        phoneme_dict["de"]["das-die"] = "DAS!DIE"
        phoneme_dict["en"]["die"] = "DAI"

        phonemizer = Phonemizer(predictor=predictor, lang_phoneme_dict=phoneme_dict)

        result = phonemizer("die", lang="de")
        self.assertEqual("DIE", result)

        result = phonemizer("Die", lang="de")
        self.assertEqual("DIE", result)

        result = phonemizer("DIE", lang="de")
        self.assertEqual("DIE", result)

        # test whether the dict check gets hyphenated words
        result = phonemizer("das-die", lang="de")
        self.assertEqual("DAS!DIE", result)

        # test whether the dict check is applied to subwords in a hyphenated word
        result = phonemizer("die-das", lang="de")
        self.assertEqual("DIE-DAS", result)

        result = phonemizer("/die/?!", lang="de", punctuation="")
        self.assertEqual("DIE", result)

        result = phonemizer("/die/?!", lang="de", punctuation="/")
        self.assertEqual("/DIE/", result)

        result = phonemizer("die", lang="en")
        self.assertEqual("DAI", result)

        result = phonemizer("die-die-das", lang="en")
        self.assertEqual("DAI-DAI-das-phon-en", result)

        result = phonemizer("dies", lang="de")
        self.assertEqual("dies-phon-de", result)

        result = phonemizer("dies", lang="en")
        self.assertEqual("dies-phon-en", result)

    @patch.object(Predictor, "__call__", new_callable=PredictorMock)
    def test_split_result(self, predictor: Predictor) -> None:
        phonemizer = Phonemizer(predictor=predictor)

        text = "Ich, bin - so wie (ich/du) bin."
        result = phonemizer.phonemise_list([text], lang="de", punctuation=",().")
        cleaned_text = "".join(result.split_text[0])
        self.assertEqual("Ich, bin - so wie (ichdu) bin.", cleaned_text)
        self.assertEqual(len(result.split_text[0]), len(result.split_phonemes[0]))

        text = "Ich, bin: - (so."
        result = phonemizer.phonemise_list([text], lang="de", punctuation=",().")
        self.assertEqual(
            ["Ich", ",", " ", "bin", " ", "-", " ", "(", "so", "."],
            result.split_text[0],
        )

        # test some boundaries
        result = phonemizer("whitespace ", lang="de")
        self.assertEqual("whitespace-phon-de ", result)

        # test some boundaries
        result = phonemizer(",special!?", lang="de", punctuation=",!")
        self.assertEqual(",special-phon-de!", result)

        # test some boundaries
        result = phonemizer(",special!?", lang="de", punctuation="")
        self.assertEqual("special-phon-de", result)

    def test_expand_acronym(self) -> None:
        expanded = Phonemizer._expand_acronym("hallo")
        self.assertEqual("hallo", expanded)

        expanded = Phonemizer._expand_acronym("HAL")
        self.assertEqual("H-A-L", expanded)

        expanded = Phonemizer._expand_acronym("hallO")
        self.assertEqual("hall-O", expanded)

        expanded = Phonemizer._expand_acronym("HaLO")
        self.assertEqual("Ha-L-O", expanded)

        expanded = Phonemizer._expand_acronym("vCPU")
        self.assertEqual("v-C-P-U", expanded)



import math
from typing import Tuple

import torch
from torch.nn.utils.rnn import pad_sequence


class PositionalEncoding(torch.nn.Module):

    def __init__(self, d_model: int, dropout=0.1, max_len=5000) -> None:
        """
        Initializes positional encoding.

        Args:
            d_model (int): Dimension of model.
            dropout (float): Dropout after positional encoding.
            max_len: Max length of precalculated position sequence.
        """

        super().__init__()
        self.dropout = torch.nn.Dropout(p=dropout)
        self.scale = torch.nn.Parameter(torch.ones(1))

        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, d_model, 2).float() * (-math.log(10000.0) / d_model)
        )
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0).transpose(0, 1)
        self.register_parameter("pe", torch.nn.Parameter(pe, requires_grad=False))

    def forward(self, x: torch.Tensor) -> torch.Tensor:  # shape: [T, N]
        x = x + self.scale * self.pe[: x.size(0), :]
        return self.dropout(x)


def get_dedup_tokens(logits_batch: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
    """Converts a batch of logits into the batch most probable tokens and their probabilities.

    Args:
      logits_batch (Tensor): Batch of logits (N x T x V).

    Returns:
      Tuple: Deduplicated tokens. The first element is a tensor (token indices) and the second element
      is a tensor (token probabilities)

    """

    logits_batch = logits_batch.softmax(-1)
    out_tokens, out_probs = [], []
    for i in range(logits_batch.size(0)):
        logits = logits_batch[i]
        max_logits, max_indices = torch.max(logits, dim=-1)
        max_logits = max_logits[max_indices != 0]
        max_indices = max_indices[max_indices != 0]
        cons_tokens, counts = torch.unique_consecutive(max_indices, return_counts=True)
        out_probs_i = torch.zeros(len(counts), device=logits.device)
        ind = 0
        for i, c in enumerate(counts):
            max_logit = max_logits[ind : ind + c].max()
            out_probs_i[i] = max_logit
            ind = ind + c
        out_tokens.append(cons_tokens)
        out_probs.append(out_probs_i)

    out_tokens = pad_sequence(out_tokens, batch_first=True, padding_value=0.0).long()
    out_probs = pad_sequence(out_probs, batch_first=True, padding_value=0.0)

    return out_tokens, out_probs


def _generate_square_subsequent_mask(sz: int) -> torch.Tensor:
    mask = torch.triu(torch.ones(sz, sz), 1)
    mask = mask.masked_fill(mask == 1, float("-inf"))
    return mask


def _make_len_mask(inp: torch.Tensor) -> torch.Tensor:
    return (inp == 0).transpose(0, 1)


def _get_len_util_stop(sequence: torch.Tensor, end_index: int) -> int:
    for i, val in enumerate(sequence):
        if val == end_index:
            return i + 1
    return len(sequence)


def _trim_util_stop(sequence: torch.Tensor, end_index: int) -> torch.Tensor:
    seq_len = _get_len_util_stop(sequence, end_index)
    return sequence[:seq_len]



from typing import Dict

import torch


class CrossEntropyLoss(torch.nn.Module):
    """ """

    def __init__(self) -> None:
        super().__init__()
        self.criterion = torch.nn.CrossEntropyLoss(ignore_index=0)

    def forward(
        self, pred: torch.Tensor, batch: Dict[str, torch.Tensor]
    ) -> torch.Tensor:
        """Forward pass of the CrossEntropyLoss module on a batch.

        Args:
          pred: Batch of model predictions.
          batch: Dictionary of a training data batch, containing 'phonemes': target phonemes.
          pred: torch.Tensor:
          batch: Dict[str:
          torch.Tensor]:

        Returns:
          Loss as tensor.

        """

        phonemes = batch["phonemes"]
        loss = self.criterion(pred.transpose(1, 2), phonemes[:, 1:])
        return loss


class CTCLoss(torch.nn.Module):
    """ """

    def __init__(self):
        super().__init__()
        self.criterion = torch.nn.CTCLoss()

    def forward(
        self, pred: torch.Tensor, batch: Dict[str, torch.Tensor]
    ) -> torch.Tensor:
        """Forward pass of the CTCLoss module on a batch.

        Args:
          pred: Batch of model predictions.
          batch: Dictionary of a training data batch, containing 'phonemes': target phonemes,
        'text_len': input text lengths, 'phonemes_len': target phoneme lengths
          pred: torch.Tensor:
          batch: Dict[str:
          torch.Tensor]:

        Returns:
          Loss as tensor.

        """

        pred = pred.transpose(0, 1).log_softmax(2)
        phonemes = batch["phonemes"]
        text_len = batch["text_len"]
        phon_len = batch["phonemes_len"]
        loss = self.criterion(pred, phonemes, text_len, phon_len)
        return loss



import unittest

import torch

from deep_phonemizer.training.dataset import PhonemizerDataset, collate_dataset


class TestPhonemizerDataset(unittest.TestCase):

    def test_get_items(self) -> None:
        items_input = [(0, [1, 2], [3, 4]), (1, [2, 3], [4, 5, 6])]
        items = PhonemizerDataset(items=items_input)

        self.assertEqual(0, items[0]["item_id"])
        self.assertEqual(0, items[0]["language"])
        self.assertEqual([1, 2], items[0]["text"].tolist())
        self.assertEqual([3, 4], items[0]["phonemes"].tolist())
        self.assertEqual(2, items[0]["text_len"])
        self.assertEqual(2, items[0]["phonemes_len"])
        self.assertEqual(3, items[0]["start_index"])

        self.assertEqual(1, items[1]["item_id"])
        self.assertEqual(1, items[1]["language"])

    def test_collate(self):
        batch = [
            {
                "item_id": 0,
                "language": 0,
                "text_len": 2,
                "phonemes_len": 2,
                "start_index": 2,
                "text": torch.tensor([1, 2]),
                "phonemes": torch.tensor([2, 3]),
            },
            {
                "item_id": 1,
                "language": 1,
                "text_len": 3,
                "phonemes_len": 3,
                "start_index": 3,
                "text": torch.tensor([1, 2, 3]),
                "phonemes": torch.tensor([3, 4, 5]),
            },
        ]

        batch = collate_dataset(batch)
        self.assertEqual(0, batch["item_id"][0])
        self.assertEqual(1, batch["item_id"][1])
        self.assertEqual(2, batch["text_len"][0])
        self.assertEqual(3, batch["text_len"][1])
        self.assertEqual(2, batch["phonemes_len"][0])
        self.assertEqual(3, batch["phonemes_len"][1])
        self.assertEqual(0, batch["language"][0])
        self.assertEqual(1, batch["language"][1])
        self.assertEqual([1, 2, 0], batch["text"][0].tolist())
        self.assertEqual([1, 2, 3], batch["text"][1].tolist())
        self.assertEqual([2, 3, 0], batch["phonemes"][0].tolist())
        self.assertEqual([3, 4, 5], batch["phonemes"][1].tolist())



import unittest

from deep_phonemizer.training.metrics import phoneme_error, word_error


class TestWordError(unittest.TestCase):

    def test_call(self):
        predicted = ["a", "b", "c", "d"]
        target = ["a", "k", "c", "a"]
        result = word_error(predicted, target)
        self.assertEqual(1, result)

        predicted = ["r", "r", "r", "r"]
        target = ["a", "k", "c", "a"]
        result = word_error(predicted, target)
        self.assertEqual(1.0, result)

        predicted = ["a"]
        target = ["a"]
        result = word_error(predicted, target)
        self.assertEqual(0, result)


class TestPhonemeErrorRate(unittest.TestCase):

    def test_call(self):
        predicted = ["a", "b", "c", "d"]
        target = ["a", "k", "c", "a"]
        e, c = phoneme_error(predicted, target)
        self.assertEqual(2, e)
        self.assertEqual(4, c)

        predicted = ["r", "r", "r", "r"]
        target = ["a", "k", "c", "a"]
        e, c = phoneme_error(predicted, target)
        self.assertEqual(4, e)
        self.assertEqual(4, c)

        predicted = ["a"]
        target = ["a"]
        e, c = phoneme_error(predicted, target)
        self.assertEqual(0, e)
        self.assertEqual(1, c)

        predicted = ["a"]
        target = ["b"]
        e, c = phoneme_error(predicted, target)
        self.assertEqual(1, e)
        self.assertEqual(1, c)

        predicted = ["a", "b"]
        target = ["b"]
        e, c = phoneme_error(predicted, target)
        self.assertEqual(1, e)
        self.assertEqual(1, c)

        predicted = ["a", "b", "c"]
        target = ["b"]
        e, c = phoneme_error(predicted, target)
        self.assertEqual(2, e)
        self.assertEqual(1, c)

        predicted = ["a"]
        target = ["a", "b"]
        e, c = phoneme_error(predicted, target)
        self.assertEqual(1, e)
        self.assertEqual(2, c)



import re
from itertools import zip_longest
from typing import Dict, List, Set, Union

from deep_phonemizer import PhonemizerResult
from deep_phonemizer.model.model import load_checkpoint
from deep_phonemizer.model.predictor import Predictor
from deep_phonemizer.utils.logging import get_logger

DEFAULT_PUNCTUATION = "().,:?!/–"


class Phonemizer:

    def __init__(
        self, predictor: Predictor, lang_phoneme_dict: Dict[str, Dict[str, str]] = None
    ) -> None:
        """
        Initializes a phonemizer with a ready predictor.

        Args:
            predictor (Predictor): Predictor object carrying the trained transformer model.
            lang_phoneme_dict (Dict[str, Dict[str, str]], optional): Word-phoneme dictionary for each language.
        """

        self.predictor = predictor
        self.lang_phoneme_dict = lang_phoneme_dict

    def __call__(
        self,
        text: Union[str, List[str]],
        lang: str,
        punctuation: str = DEFAULT_PUNCTUATION,
        expand_acronyms: bool = True,
        batch_size: int = 8,
    ) -> Union[str, List[str]]:
        """
        Phonemizes a single text or list of texts.

        Args:
          text (str): Text to phonemize as single string or list of strings.
          lang (str): Language used for phonemization.
          punctuation (str): Punctuation symbols by which the texts are split.
          expand_acronyms (bool): Whether to expand an acronym, e.g. DIY -> D-I-Y.
          batch_size (int): Batch size of model to speed up inference.

        Returns:
          Union[str, List[str]]: Phonemized text as string, or list of strings, respectively.
        """

        single_input_string = isinstance(text, str)
        texts = [text] if single_input_string else text
        result = self.phonemise_list(
            texts=texts,
            lang=lang,
            punctuation=punctuation,
            expand_acronyms=expand_acronyms,
        )

        phoneme_lists = ["".join(phoneme_list) for phoneme_list in result.phonemes]

        if single_input_string:
            return phoneme_lists[0]
        else:
            return phoneme_lists

    def phonemise_list(
        self,
        texts: List[str],
        lang: str,
        punctuation: str = DEFAULT_PUNCTUATION,
        expand_acronyms: bool = True,
        batch_size: int = 8,
    ) -> PhonemizerResult:
        """Phonemizes a list of texts and returns tokenized texts,
        phonemes and word predictions with probabilities.

        Args:
          texts (List[str]): List texts to phonemize.
          lang (str): Language used for phonemization.
          punctuation (str): Punctuation symbols by which the texts are split. (Default value = DEFAULT_PUNCTUATION)
          expand_acronyms (bool): Whether to expand an acronym, e.g. DIY -> D-I-Y. (Default value = True)
          batch_size (int): Batch size of model to speed up inference. (Default value = 8)

        Returns:
          PhonemizerResult: Object containing original texts, phonemes, split texts, split phonemes, and predictions.

        """

        punc_set = set(punctuation + "- ")
        punc_pattern = re.compile(f'([{punctuation + " "}])')

        split_text, cleaned_words = [], set()
        for text in texts:
            # cleaned_text = ''.join([t for t in text if t.isalnum() or t in punc_set])
            cleaned_text = text
            split = re.split(punc_pattern, cleaned_text)
            split = [s for s in split if len(s) > 0]
            split_text.append(split)
            cleaned_words.update(split)

        # collect dictionary phonemes for words and hyphenated words
        word_phonemes = {
            word: self._get_dict_entry(word=word, lang=lang, punc_set=punc_set)
            for word in cleaned_words
        }

        # if word is not in dictionary, split it into subwords
        # words_to_split = [w for w in cleaned_words if word_phonemes[w] is None]
        words_to_split = []
        word_splits = dict()
        for word in words_to_split:
            key = word
            word = self._expand_acronym(word) if expand_acronyms else word
            word_split = re.split(r"([-])", word)
            word_splits[key] = word_split

        # collect dictionary entries of subwords
        subwords = {w for values in word_splits.values() for w in values}
        subwords = {w for w in subwords if w not in word_phonemes}
        for subword in subwords:
            word_phonemes[subword] = self._get_dict_entry(
                word=subword, lang=lang, punc_set=punc_set
            )

        # predict all subwords that are missing in the phoneme dict
        words_to_predict = [
            word
            for word, phons in word_phonemes.items()
            if phons is None and len(word_splits.get(word, [])) <= 1
        ]

        predictions = self.predictor(
            words=words_to_predict, lang=lang, batch_size=batch_size
        )

        word_phonemes.update({pred.word: pred.phonemes for pred in predictions})
        pred_dict = {pred.word: pred for pred in predictions}

        # collect all phonemes
        phoneme_lists = []
        for text in split_text:
            text_phons = [
                self._get_phonemes(
                    word=word, word_phonemes=word_phonemes, word_splits=word_splits
                )
                for word in text
            ]
            phoneme_lists.append(text_phons)

        phonemes_joined = ["".join(phoneme_list) for phoneme_list in phoneme_lists]

        return PhonemizerResult(
            text=texts,
            phonemes=phonemes_joined,
            split_text=split_text,
            split_phonemes=phoneme_lists,
            predictions=pred_dict,
        )

    def _get_dict_entry(
        self, word: str, lang: str, punc_set: Set[str]
    ) -> Union[str, None]:
        if word in punc_set or len(word) == 0:
            return word
        if not self.lang_phoneme_dict or lang not in self.lang_phoneme_dict:
            return None
        phoneme_dict = self.lang_phoneme_dict[lang]
        if word in phoneme_dict:
            return phoneme_dict[word]
        elif word.lower() in phoneme_dict:
            return phoneme_dict[word.lower()]
        elif word.title() in phoneme_dict:
            return phoneme_dict[word.title()]
        else:
            return None

    @staticmethod
    def _expand_acronym(word: str) -> str:
        subwords = []
        for subword in word.split("-"):
            expanded = []
            for a, b in zip_longest(subword, subword[1:]):
                expanded.append(a)
                if b is not None and b.isupper():
                    expanded.append("-")
            expanded = "".join(expanded)
            subwords.append(expanded)
        return "-".join(subwords)

    @staticmethod
    def _get_phonemes(
        word: str,
        word_phonemes: Dict[str, Union[str, None]],
        word_splits: Dict[str, List[str]],
    ) -> str:
        phons = word_phonemes[word]
        if phons is None:
            subwords = word_splits[word]
            subphons = [word_phonemes[w] for w in subwords]
            phons = "".join(subphons)
        return phons

    @classmethod
    def from_checkpoint(
        cls,
        checkpoint_path: str,
        device="cpu",
        lang_phoneme_dict: Dict[str, Dict[str, str]] = None,
    ) -> "Phonemizer":
        """Initializes a Phonemizer object from a model checkpoint (.pt file).

        Args:
          checkpoint_path (str): Path to the .pt checkpoint file.
          device (str): Device to send the model to ('cpu' or 'cuda'). (Default value = 'cpu')
          lang_phoneme_dict (Dict[str, Dict[str, str]], optional): Word-phoneme dictionary for each language.

        Returns:
          Phonemizer: Phonemizer object carrying the loaded model and, optionally, a phoneme dictionary.
        """

        model, checkpoint = load_checkpoint(checkpoint_path, device=device)
        applied_phoneme_dict = None
        if lang_phoneme_dict is not None:
            applied_phoneme_dict = lang_phoneme_dict
        elif "phoneme_dict" in checkpoint:
            applied_phoneme_dict = checkpoint["phoneme_dict"]
        preprocessor = checkpoint["preprocessor"]
        predictor = Predictor(model=model, preprocessor=preprocessor)
        logger = get_logger(__name__)
        model_step = checkpoint["step"]
        logger.debug(f"Initializing phonemizer with model step {model_step}")
        return Phonemizer(predictor=predictor, lang_phoneme_dict=applied_phoneme_dict)



from typing import Any, Dict, List, Tuple

from deep_phonemizer.training.metrics import phoneme_error, word_error


def evaluate_samples(
    lang_samples: Dict[str, List[Tuple[List[str], List[str], List[str]]]]
) -> Dict[str, Any]:
    """Calculates word and phoneme error rates per language and their mean across languages

    Args:
      lang_samples (Dict): Data to evaluate. Contains languages as keys and list of result samples as values.
                           Prediction samples is given as a List of Tuples, where each Tuple is a tokenized representation of
                           (text, result, target).

    Returns:
      Dict: Evaluation result carrying word and phoneme error rates per language.

    """

    evaluation_result = dict()
    lang_phon_err, lang_phon_count, lang_word_err = dict(), dict(), dict()
    languages = sorted(lang_samples.keys())
    for lang in languages:
        for word, generated, target in lang_samples[lang]:
            word = "".join(word)
            phon_err, phon_count = phoneme_error(generated, target)
            word_err = word_error(generated, target)
            phon_err_dict = lang_phon_err.setdefault(lang, dict())
            phon_count_dict = lang_phon_count.setdefault(lang, dict())
            word_err_dict = lang_word_err.setdefault(lang, dict())
            best_phon_err, best_phon_count = phon_err_dict.get(
                word, None
            ), phon_count_dict.get(word, None)
            if (
                best_phon_err is None
                or phon_err / phon_count < best_phon_err / best_phon_count
            ):
                phon_err_dict[word] = phon_err
                phon_count_dict[word] = phon_count
                word_err_dict[word] = word_err

    phon_errors, phon_counts, word_errors, word_counts = [], [], [], []
    for lang in languages:
        phon_err = sum(lang_phon_err[lang].values())
        phon_errors.append(phon_err)
        phon_count = sum(lang_phon_count[lang].values())
        phon_counts.append(phon_count)
        word_err = sum(lang_word_err[lang].values())
        word_errors.append(word_err)
        word_count = len(lang_word_err[lang])
        word_counts.append(word_count)
        per = phon_err / phon_count
        wer = word_err / word_count
        evaluation_result.setdefault(lang, {}).update({"per": per})
        evaluation_result.setdefault(lang, {}).update({"wer": wer})
    mean_per = sum(phon_errors) / sum(phon_counts)
    mean_wer = sum(word_errors) / sum(word_counts)
    evaluation_result["mean_per"] = mean_per
    evaluation_result["mean_wer"] = mean_wer

    return evaluation_result



from typing import Dict, List, Tuple

import torch
from torch.nn.utils.rnn import pad_sequence

from deep_phonemizer import Prediction
from deep_phonemizer.model.model import load_checkpoint
from deep_phonemizer.model.utils import _get_len_util_stop
from deep_phonemizer.preprocessing.text import Preprocessor
from deep_phonemizer.preprocessing.utils import _batchify, _product


class Predictor:
    """Performs model predictions on a batch of inputs."""

    def __init__(self, model: torch.nn.Module, preprocessor: Preprocessor) -> None:
        """
        Initializes a Predictor object with a trained transformer model a preprocessor.

        Args:
            model (Model): Trained transformer model.
            preprocessor (Preprocessor): Preprocessor corresponding to the model configuration.
        """

        self.model = model
        self.text_tokenizer = preprocessor.text_tokenizer
        self.phoneme_tokenizer = preprocessor.phoneme_tokenizer

    def __call__(
        self, words: List[str], lang: str, batch_size: int = 8
    ) -> List[Prediction]:
        """
        Predicts phonemes for a list of words.

        Args:
          words (list): List of words to predict.
          lang (str): Language of texts.
          batch_size (int): Size of batch for model input to speed up inference.

        Returns:
          List[Prediction]: A list of result objects containing (word, phonemes, phoneme_tokens, token_probs, confidence)
        """

        predictions = dict()
        valid_texts = set()

        # handle words that result in an empty input to the model
        for word in words:
            input = self.text_tokenizer(sentence=word, language=lang)
            decoded = self.text_tokenizer.decode(
                sequence=input, remove_special_tokens=True
            )
            if len(decoded) == 0:
                predictions[word] = ([], [])
            else:
                valid_texts.add(word)

        valid_texts = sorted(list(valid_texts), key=lambda x: len(x))
        batch_pred = self._predict_batch(
            texts=valid_texts, batch_size=batch_size, language=lang
        )
        predictions.update(batch_pred)

        output = []
        for word in words:
            tokens, probs = predictions[word]
            out_phons = self.phoneme_tokenizer.decode(
                sequence=tokens, remove_special_tokens=True
            )
            out_phons_tokens = self.phoneme_tokenizer.decode(
                sequence=tokens, remove_special_tokens=False
            )
            output.append(
                Prediction(
                    word=word,
                    phonemes="".join(out_phons),
                    phoneme_tokens=out_phons_tokens,
                    confidence=_product(probs),
                    token_probs=probs,
                )
            )

        return output

    def _predict_batch(
        self, texts: List[str], batch_size: int, language: str
    ) -> Dict[str, Tuple[List[int], List[float]]]:
        """
        Returns dictionary with key = word and val = Tuple of (phoneme tokens, phoneme probs)
        """

        predictions = dict()
        text_batches = _batchify(texts, batch_size)
        for text_batch in text_batches:
            input_batch, lens_batch = [], []
            for text in text_batch:
                input = self.text_tokenizer(text, language)
                input_batch.append(torch.tensor(input))
                lens_batch.append(torch.tensor(len(input)))

            input_batch = pad_sequence(
                sequences=input_batch, batch_first=True, padding_value=0
            )
            lens_batch = torch.stack(lens_batch)
            start_indx = self.phoneme_tokenizer._get_start_index(language)
            start_inds = torch.tensor([start_indx] * input_batch.size(0)).to(
                input_batch.device
            )
            batch = {
                "text": input_batch,
                "text_len": lens_batch,
                "start_index": start_inds,
            }
            with torch.no_grad():
                output_batch, probs_batch = self.model.generate(batch)
            output_batch, probs_batch = output_batch.cpu(), probs_batch.cpu()
            for text, output, probs in zip(text_batch, output_batch, probs_batch):
                seq_len = _get_len_util_stop(output, self.phoneme_tokenizer.end_index)
                predictions[text] = (
                    output[:seq_len].tolist(),
                    probs[:seq_len].tolist(),
                )

        return predictions

    @classmethod
    def from_checkpoint(cls, checkpoint_path: str, device="cpu") -> "Predictor":
        """Initializes the predictor from a checkpoint (.pt file).

        Args:
          checkpoint_path (str): Path to the checkpoint file (.pt).
          device (str): Device to load the model on ('cpu' or 'cuda'). (Default value = 'cpu').

        Returns:
          Predictor: Predictor object.

        """
        model, checkpoint = load_checkpoint(checkpoint_path, device=device)
        preprocessor = checkpoint["preprocessor"]
        return Predictor(model=model, preprocessor=preprocessor)



from typing import Dict, List


class Prediction:
    """
    Container for single word prediction result.
    """

    def __init__(
        self,
        word: str,
        phonemes: str,
        phoneme_tokens: List[str],
        confidence: float,
        token_probs: List[float],
    ) -> None:
        """
        Initializes a Prediction object.

        Args:
          word (str): Original word to predict.
          phonemes (str): Predicted phonemes (without special tokens).
          phoneme_tokens (List[str]): Predicted phoneme tokens (including special tokens).
          confidence (float): Total confidence of result.
          token_probs (List[float]): Probability of each phoneme token.
        """

        self.word = word
        self.phonemes = phonemes
        self.phoneme_tokens = phoneme_tokens
        self.confidence = confidence
        self.token_probs = token_probs


class PhonemizerResult:
    """
    Container for phonemizer output.
    """

    def __init__(
        self,
        text: List[str],
        phonemes: List[str],
        split_text: List[List[str]],
        split_phonemes: List[List[str]],
        predictions: Dict[str, Prediction],
    ) -> None:
        """
        Initializes a PhonemizerResult object.

        Args:
          text (List[str]): List of input texts.
          phonemes (List[str]): List of output phonemes.
          split_text (List[List[str]]): List of texts, where each text is split into words and special chars.
          split_phonemes (List[List[str]]): List of phonemes corresponding to split_text.
          predictions (Dict[str, Prediction]): Dictionary with entries word to Tuple (phoneme, probability).
        """

        self.text = text
        self.phonemes = phonemes
        self.split_text = split_text
        self.split_phonemes = split_phonemes
        self.predictions = predictions



import unittest

from deep_phonemizer.training.evaluation import evaluate_samples


class TestEvaluation(unittest.TestCase):

    def test_evaluate_samples(self):

        lang_samples = {
            "de": [
                (["a"], ["a", "b"], ["a", "b"]),
            ],
            "en_us": [
                # two alternative targets, evaluation should take the closer one with per=1./3
                (["a"], ["a", "b", "c"], ["b", "b", "b"]),
                (["a"], ["a", "b", "c"], ["a", "b", "b"]),
            ],
        }

        result = evaluate_samples(lang_samples=lang_samples)

        self.assertEqual(0.0, result["de"]["per"])
        self.assertEqual(0.0, result["de"]["wer"])
        self.assertEqual(1.0 / 3, result["en_us"]["per"])
        self.assertEqual(1.0, result["en_us"]["wer"])
        self.assertEqual((0 + 1) / (2 + 3), result["mean_per"])
        self.assertEqual((0 + 1) / 2, result["mean_wer"])



import unittest

from deep_phonemizer.preprocessing.text import SequenceTokenizer


class TestSequenceTokenizer(unittest.TestCase):

    def test_call_happy_path(self) -> None:
        symbols = ["a", "b", "c", "d", "e"]
        languages = ["de", "en"]

        tokenizer = SequenceTokenizer(
            symbols=symbols,
            languages=languages,
            char_repeats=1,
            lowercase=True,
            append_start_end=True,
            end_token="<end>",
        )

        tokens = tokenizer(["a", "b"], language="de")
        decoded = tokenizer.decode(tokens)

        self.assertEqual(4, len(tokens))
        self.assertEqual(["<de>", "a", "b", "<end>"], decoded)

        tokens = tokenizer(["a", "b"], language="en")
        decoded = tokenizer.decode(tokens)

        self.assertEqual(4, len(tokens))
        self.assertEqual(["<en>", "a", "b", "<end>"], decoded)

    def test_call_missing_symbols(self) -> None:
        symbols = ["a", "b"]
        languages = ["de", "en"]

        tokenizer = SequenceTokenizer(
            symbols=symbols,
            languages=languages,
            lowercase=False,
            append_start_end=True,
            end_token="<end>",
            char_repeats=2,
        )

        tokens = tokenizer(["A", "b", "F"], language="en")
        decoded = tokenizer.decode(tokens)

        self.assertEqual(4, len(tokens))
        self.assertEqual(["<en>", "b", "<end>"], decoded)

    def test_call_edge_cases(self) -> None:
        symbols = ["a", "b", "c", "d", "e"]
        languages = ["de", "en"]

        tokenizer = SequenceTokenizer(
            symbols=symbols,
            languages=languages,
            lowercase=True,
            append_start_end=True,
            end_token="<end>",
            char_repeats=2,
        )

        tokens = tokenizer([], language="en")
        decoded = tokenizer.decode(tokens)

        self.assertEqual(2, len(tokens))
        self.assertEqual(["<en>", "<end>"], decoded)

        tokens = tokenizer(["z"], language="en")
        decoded = tokenizer.decode(tokens)

        self.assertEqual(2, len(tokens))
        self.assertEqual(["<en>", "<end>"], decoded)

        tokenizer = SequenceTokenizer(
            symbols=["a"],
            languages=["de"],
            lowercase=True,
            append_start_end=False,
            end_token=None,
            char_repeats=3,
        )

        tokens = tokenizer(["a"], language="de")
        decoded = tokenizer.decode(tokens)

        self.assertEqual(3, len(tokens))
        self.assertEqual(["a"], decoded)

    def test_exception(self) -> None:
        tokenizer = SequenceTokenizer(
            symbols=["a"],
            languages=["de"],
            lowercase=True,
            append_start_end=True,
            end_token="<end>",
            char_repeats=2,
        )

        with self.assertRaises(ValueError):
            tokenizer(["a", "b"], language="not_existent")
